import importlib
import json
import os
import re
import tempfile
import typing
from hashlib import sha256
from pathlib import Path
from typing import Type

import pydantic_yaml as pydml
from datamodel_code_generator import generate
from deepdiff import DeepDiff
from jinja2 import Environment, FileSystemLoader
from pydantic import BaseModel, Field

import malevich
from malevich.constants import reserved_config_fields
from malevich.core_api import AppFunctionsInfo, ConditionFunctionInfo
from malevich.models.dependency import Dependency

from ..path import Paths


class Templates:
    """Templates for creating a package"""

    module_general_description = """\"\"\"
Third-party Malevich app. Use it in flow by importing any processor
from this module and calling it within @flow decorated function.
\"\"\"
"""

    module_specific_description = """\"\"\"
{description}

---

Third-party Malevich app. Use it in flow by importing any processor
from this module and calling it within @flow decorated function.
\"\"\"
"""

    imports = """
import typing
from typing import *

import malevich.annotations
from malevich.models import ConfigArgument
from malevich._meta.decor import proc
from malevich._utility import Registry
from malevich.models.nodes import OperationNode
from .scheme import *
"""

    registry = """
Registry().register("{operation_id}", {registry_record})
"""

    processor = """
{definition}

{processor_name} = proc(use_sinktrace={use_sinktrace}, config_model={config_model})(__{processor_name})
\"""{docstrings}\"""
"""  # noqa: E501

def create_package_stub(
    package_name: str,
    metascript: str,
    dependency: Dependency
) -> tuple[bool, str]:
    """
    Create a package with the given name and contents
    Args:
        package (str): The name of the package to create.
        metascript (str): The contents of the package.

    Returns:
        A tuple containing a boolean indicating whether the package was
        created successfully, and the checksum of the package.
    """
    # Calculate the checksum of the package contents
    checksum = sha256(metascript.encode()).hexdigest()

    # Create the package directory
    # The package will be located at <python>/malevich/<package>
    package_path = Path(malevich.__file__).parent / package_name
    # Create the package directory if it doesn't exist
    os.makedirs(package_path, exist_ok=True)

    # Create the package __init__.py file
    with open(package_path / "__init__.py", "w") as f:
        # Write the metascript to the file
        f.write(metascript)
        # Write the checksum to the file
        f.write(f"\n__Metascript_checksum__ = '{checksum}'\n")
        f.write(f"\n__Dependency_checksum__ = '{dependency.checksum()}'\n")

    # Post-hoc check to make sure the package was created successfully
    try:
        # Import the package
        __pkg = importlib.import_module(f"malevich.{package_name}")
        # Assert that the checksum of the package is the same as the checksum
        assert getattr(__pkg, "__Metascript_checksum__") == checksum, \
            "Package checksum does not match metascript checksum. "
    except (ImportError, AssertionError):
        return False, checksum
    return True, checksum


class StubFunction(BaseModel):
    name: str
    # (name, type,)
    args: list[tuple[str, str | None]]
    # (name, pos,)
    sink: tuple[str, int] | None
    docstrings: str | None = None
    config_schema: Type[BaseModel] | None = Field(None, exclude=True)
    definition: str | None = None
    package_id: str
    is_condition: bool = False
    operation_id: str

    def generate_definition(self, config_model: str | None = None) -> str:
        """def processor_name(
            arg1: type1,
            arg2: type2,
            *,
            var1: type1,
            var2: type2 | None = None
            config: ConfigModel,
            **kwargs
        ) -> RETURN_TYPE:
        """
        environment = Environment(
            loader=FileSystemLoader(Paths.templates('metascript', 'op_stub'))
        )
        extra_kwargs_ = {}
        if self.config_schema:
            config_fields = self.config_schema.model_fields
            for field in config_fields:
                if field == "config":
                    raise ValueError(
                        "Trying to generate a definition for "
                        "a function with a config field named 'config'"
                    )
                annotation = config_fields[field].annotation
                if (
                    hasattr(annotation, "__name__")
                    and annotation.__module__ != typing.__name__
                ):
                    # class
                    annotation = config_fields[field].annotation.__name__
                else:
                    # etc.
                    annotation = str(config_fields[field].annotation)
                config_fields[field].annotation = annotation

            extra_kwargs_['config_fields'] = config_fields
        if self.sink:
            extra_kwargs_['sink'] = self.sink

        data = environment.get_template('def.jinja2').render(
            processor_name = self.name,
            args = self.args,
            docstrings = self.docstrings,
            is_condition = self.is_condition,
            package_id = self.package_id,
            config_model=config_model,
            reserved_config_fields=reserved_config_fields,
            operation_id=self.operation_id,
            **extra_kwargs_
        )

        return data


class StubSchema(BaseModel):
    name: str
    scheme: str
    class_name: str


class StubIndex(BaseModel):
    dependency: Dependency
    functions: list[StubFunction]
    functions_index: dict[str, tuple[int, int]]
    schemes: list[StubSchema]
    schemes_index: dict[str, tuple[int, int]]


class Stub:
    class Utils:
        @staticmethod
        def combine_schemas(schemas: list[str]) -> str:
            schemas = [json.loads(schema) for schema in schemas]
            all_defs = {}
            for schema in schemas:
                if not schema:
                    continue

                defs = schema.get('$defs', schema.get('definitions', {}))
                for def_name, def_schema in defs.items():
                    if def_name in all_defs:
                        diff = DeepDiff(all_defs[def_name], def_schema)
                        if diff:
                            raise ValueError(
                                "Duplicate schemas found with different content: "
                                f"{def_name}"
                            )
                    else:
                        all_defs[def_name] = def_schema

                if 'title' in schema:
                    all_defs[schema['title']] = schema

            return json.dumps({
                "$defs": all_defs,
                "type": "object",
            })


        @staticmethod
        def generate_context_schema(json_schema: str) -> tuple[str, str]:
            with (
                tempfile.NamedTemporaryFile(mode='w+', suffix='.json') as f,
                tempfile.NamedTemporaryFile(mode='w+', suffix='.py') as out
            ):
                f.write(json_schema)
                f.seek(0)
                generate(
                    Path(f.name),
                    output=Path(out.name),
                    use_annotated=False,
                    input_file_type='jsonschema',
                    base_class='malevich.models._model._Model', # cool
                    collapse_root_models=True,
                    disable_timestamp=True,
                    custom_file_header='"""Malevich auto-generated schema"""',
                )
                out_script =  open(out.name).read().replace(
                    'from __future__ import annotations',
                    ''
                )
                return (
                    re.findall(
                        r'class (?P<ClassName>\w+)',
                        out_script
                    ),   # 1
                    out_script              # 2
                )


    def __init__(
        self,
        path: str,
        malevich_app_name: str | None = None,
        dependency: Dependency | None = None
    ) -> None:
        ipath = os.path.join(path, 'index.yaml')
        assert Path(ipath).exists(), "No index.yaml found in the package"
        self.index = pydml.parse_yaml_file_as(StubIndex, ipath)
        self.path = path
        self.malevich_app_name = malevich_app_name
        self.dependency = dependency

    @staticmethod
    def from_app_info(
        path: str,
        package_name: str,
        dependency: Dependency,
        app_info: AppFunctionsInfo,
        operation_ids: dict[str, str],
        registry_records: dict[str, dict],
        description: str | None = None,
    ) -> "Stub":
        os.makedirs(path, exist_ok=True)
        operations = {**app_info.processors, **app_info.conditions}
        operations = {str(key): value for key, value in operations.items()}

        index = StubIndex(
            dependency=dependency,
            functions=[],
            functions_index={},
            schemes=[],
            schemes_index={},
        )
        config_stubs = {
            name: Stub.Utils.generate_context_schema(
                json.dumps(processor.contextClass),
            ) if processor.contextClass is not None else (None, None)
            for name, processor in operations.items()
        }

        schemes = [
            *app_info.schemes.values(),
            *[json.dumps(processor.contextClass) for processor in operations.values()]
        ]

        config_model_class = {}
        for processor_name, (class_names, stub) in config_stubs.items():
            if class_names:
                for class_name in class_names:
                    if class_name == operations[processor_name].contextClass['title']:
                        config_model_class[processor_name] = class_name
                        break
                else:
                    config_model_class[processor_name] = None
            else:
                config_model_class[processor_name] = None


        with open(os.path.join(path, 'scheme.py'), 'w+') as f_scheme:
            f_scheme.write(
                Stub.Utils.generate_context_schema(
                    Stub.Utils.combine_schemas(schemes)
                )[1]
            )


        importlib.import_module(f'malevich.{package_name}.scheme')
        config_models = {
            name: eval(f'malevich.{package_name}.scheme.{config_model_class[name]}')
            if config_model_class[name] else None
            for name in operations
        }

        functions: dict[str, StubFunction] = {}
        schemes = {*app_info.schemes.values()}

        for name, processor in operations.items():
            args = processor.arguments
            parsed = []
            sink = None

            for idx, (alias, type_) in enumerate(args):
                if 'return' in alias:
                    continue
                if type_:
                    if 'Sink' in type_:
                        sink = (alias, idx,)
                        continue
                    if 'Context' in type_:
                        continue
                    for stub in schemes:
                        if stub in type_:
                            type_ = stub
                            break
                parsed.append((
                    alias,
                    'malevich.annotations.OpResult | malevich.annotations.Collection'
                ))

            functions[name] = StubFunction(
                name=name,
                args=parsed,
                sink=sink,
                docstrings=processor.doc,
                config_schema=config_models[name],
                package_id=package_name,
                is_condition = isinstance(processor, ConditionFunctionInfo),
                operation_id=operation_ids[name]
            )
            functions[name].definition = functions[name].generate_definition(
                config_model=config_model_class[name],
            )

        with open(os.path.join(path, 'F.py'), 'w+') as f_F:  # noqa: N806
            i = f_F.write(Templates.imports)

            for name, function in functions.items():
                if name not in operation_ids:
                    continue

                j = f_F.write(Templates.registry.format(
                    operation_id=operation_ids[name],
                    registry_record=registry_records[name]
                ))

                j += f_F.write(Templates.processor.format(
                    use_sinktrace=bool(function.sink),
                    definition=function.definition,
                    config_model=config_model_class[name],
                    processor_name=function.name,
                    docstrings=function.docstrings
                ))
                index.functions.append(function)
                index.functions_index[name] = (i, i + j)
                i += j

        pydml.to_yaml_file(os.path.join(path, 'index.yaml'), index)
        with open(os.path.join(path, '__init__.py'), 'w+') as init:
            if description:
                init.write(Templates.module_specific_description.format(
                    description=description
                ))
            else:
                init.write(Templates.module_general_description)

            init.write(
                "\n"
                "from .F import *\n"
                "from .scheme import *\n"
            )

        return Stub(
            path=path,
            malevich_app_name=package_name,
            dependency=dependency
        )



from typing import Callable, List, Optional


def input_doc(id: Optional[str] = None, collection_from: Optional[str] = None, collections_from: Optional[List[str]] = None, extra_collection_from: Optional[str] = None, extra_collections_from: Optional[List[str]] = None, query: Optional[str] = None, finish_msg: Optional[str] = None):  # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def input_df(id: Optional[str] = None, collection_from: Optional[str] = None, collections_from: Optional[List[str]] = None, extra_collection_from: Optional[str] = None, extra_collections_from: Optional[List[str]] = None, by_args: bool = False, query: Optional[str] = None, finish_msg: Optional[str] = None):  # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def input_true(id: Optional[str] = None, collection_from: Optional[str] = None, collections_from: Optional[List[str]] = None, extra_collection_from: Optional[str] = None, extra_collections_from: Optional[List[str]] = None, query: Optional[str] = None, finish_msg: Optional[str] = None):  # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def processor(id: Optional[str] = None, finish_msg: Optional[str] = None, drop_internal: bool = True, get_scale_part_all: bool = False):    # noqa: ANN201, E501
    """Denotes a processor of the app.

    Processors are core logical units of the app. To
    understand processors more, see :doc:`What is Processor? </SDK/Apps/Processors>`
    
    Args:
        id: The id of the processor. If not provided, the name of the function will be used.
        finish_msg: The message to be sent to e-mail (if configured) when the processor finishes.
        drop_internal: Whether to drop the internal collections after the processor finishes.
        get_scale_part_all: Whether to get the scale part of the input collection.

    Returns:
        The decorated function.
    """ # noqa: E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def output(id: Optional[str] = None, collection_name: Optional[str] = None, collection_names: Optional[List[str]] = None, finish_msg: Optional[str] = None, drop_internal: bool = True):    # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def condition(id: Optional[str] = None, finish_msg: Optional[str] = None, drop_internal: bool = True):    # noqa: ANN201, E501
    """Denotes a condition of the app.

    Conditions are core logical units of the app.

    Args:
        id: The id of the condition. If not provided, the name of the function will be used.
        finish_msg: The message to be sent to e-mail (if configured) when the condition finishes.
        drop_internal: Whether to drop the internal collections after the condition finishes.
        get_scale_part_all: Whether to get the scale part of the input collection.

    Returns:
        The decorated function.
    """ # noqa: E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def scheme():  # noqa: ANN201
    """Denotes a class as a scheme that might be used in DF/DFS/M/Sink annotations.

    The class automatically derives from :class:`pydantic.BaseModel`.

    Example:

    .. code-block:: python

        @scheme()
        class Users:
            name: str
            age: int

        @processor()
        def process_users(users: DF[Users]):
    """
    def wrapper(cl):  # noqa: ANN202
        return cl
    return wrapper


def init(id: Optional[str] = None, enable: bool = True, tl: Optional[int] = None, prepare: bool = False):   # noqa: ANN201, E501
    """Denotes an initialization function.

    Initialization functions are called before the app starts.
    Depending on the configuration, they might be called on
    task preparation stage or before each task run.

    Args:
        id: The id of the initialization function. If not provided, the name of the function will be used.
        enable: Whether to enable the initialization function. Defaults to True.
        tl: The time limit of the initialization function (in seconds).
        prepare: Whether to call the initialization function on task preparation stage.
    """ # noqa: E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper



from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    ParamSpec,
    TypeVar,
    overload,
)

from pydantic import BaseModel

from malevich._deploy import Space
from malevich._utility import get_space_leaves, unique
from malevich.models import Collection, SpaceCollectionResult, SpaceTask

from .._autoflow.flow import Flow
from .._autoflow.tracer import traced, tracedLike
from ..models.argument import ArgumentLink
from ..models.nodes.collection import CollectionNode
from ..models.nodes.operation import OperationNode
from ..models.nodes.tree import TreeNode

FnArgs = ParamSpec("FnArgs")
FnReturn = TypeVar("FnReturn")
Config = TypeVar("Config", bound=BaseModel)
ProcConfig = TypeVar("ProcConfig", bound=BaseModel)

ProcFunArgs = ParamSpec("ProcFunArgs")
ProcFunReturn = TypeVar("ProcFunReturn")

class IntegratedFlowStub(Generic[ProcFunArgs, ProcFunReturn]):
    def __init__(
        self,
        fn,
        reverse_id
    ) -> None:
        self._fn_ = fn
        self.reverse_id = reverse_id

    @overload
    def __call__(
        self,
        *args,
        get_task: Literal[True] = True,
        **kwargs
    ) -> SpaceTask:
        pass

    @overload
    def __call__(
        self,
        *args,
        get_task: Literal[False] = False,
        **kwargs
    ) -> list[SpaceCollectionResult]:
        pass

    @overload
    def __call__(
        self,
        *args,
        get_task: Literal[False] = False,
        wait_for_results: Literal[False] = False,
        **kwargs
    ) -> str:
        pass

    def __call__(
        self,
        version=None,
        branch=None,
        *args,
        run_deployment_id: str | None = None,
        run_task_policy: Literal['only_use', 'use_or_new', 'no_use'] = 'use_or_new',
        get_task: bool | None = None,
        wait_for_results: bool | None = None,
        webhook_url: str | None = None,
        **kwargs
    ) -> list[SpaceCollectionResult]:
        task = Space(
            reverse_id=self.reverse_id,
            branch=branch,
            version=version,
            deployment_id=run_deployment_id,
            task_policy=run_task_policy
        )

        injectables = task.get_injectables()
        expected_keys = {col.alias for col in injectables}

        if Flow.isinflow():
            if get_task is not None or wait_for_results is not None:
                raise ValueError(
                    f"You called integrated flow function {self.reverse_id} with "
                    f"get_task or wait_for_results set to a value. This is not allowed "
                    "when calling integrated flow functions from within a "
                    "@flow decorated function. If you intended to run the flow "
                    "or obtain the task, please call the function "
                    " from outside the flow."
                    # TODO: documentation ref
                )

            in_flow_connections: dict[str, traced] = {}
            flow_inputs = {}
            for col in injectables:
                data = kwargs.get(col.alias, None)
                if not isinstance(data, traced):
                    raise ValueError(
                        f"You called integrated flow function {self.reverse_id} "
                        "from within a flow function, but the argument "
                        f"{col.alias} is not a proper value. When defining a flow "
                        "with integrated subflow, you must pass an output of "
                        "Malevich operations such are processors or other flows. "
                        f"The {type(data)} value was passed instead."
                    )

                in_flow_connections[col.alias] = data
                flow_inputs[col.alias] = tracedLike(
                    CollectionNode(
                        alias=col.alias,
                        collection=Collection(collection_id=col.alias)
                    )
                )


            leaves = get_space_leaves(task.component)
            tree_node = TreeNode(
                reverse_id=self.reverse_id,
                alias=unique(self.reverse_id),
                name=task.component.name,
                description=task.component.description,
                integrated=True
            )
            tracer = traced(tree_node)
            for inp in in_flow_connections:
                in_flow_connections[inp]._autoflow.calledby(
                    tracer,
                    ArgumentLink(
                        index=0, # Collection is always the first argument
                        name=inp,
                        is_compressed_edge=True,
                        compressed_edges=[(
                            ArgumentLink(
                                index=0,
                                name=inp,
                            ),
                            flow_inputs[inp]
                        )]
                    )
                )

            traced_leaves = [
                OperationNode(
                    alias=leaf.alias,
                    operation_id=leaf.app.active_op[0].uid,
                )
                for leaf in leaves
            ]

            outputs = [
                traced(TreeNode(
                    **tree_node.model_dump(exclude=['underlying_node']),
                    underlying_node=leaf
                ))
                for leaf in traced_leaves
            ]

            return outputs

        if get_task:
            return task

        if get_task is None:
            get_task = False
        if wait_for_results is None:
            wait_for_results = True

        overrides = {}
        for col in injectables:
            data = kwargs.get(col.alias, None)
            if data is not None:
                overrides[col.alias] = data
                expected_keys.remove(col.alias)

        if expected_keys:
            for key, value in zip(expected_keys, args, strict=False):
                overrides[key] = value

        run_id = task.run(override=overrides, webhook_url=webhook_url)
        if wait_for_results:
            return task.results()
        return run_id



def integrated(reverse_id: str) -> Callable:
    def decorator(fn: Callable) -> IntegratedFlowStub[Callable[..., Any], Any]:
        return IntegratedFlowStub(fn, reverse_id)
    return decorator



import sys
from contextlib import redirect_stderr, redirect_stdout
from functools import partial, wraps
from inspect import isfunction
from io import StringIO
from typing import TypeVar

from requests.exceptions import HTTPError

from malevich._analytics import manager

T = TypeVar('T')

def _func_name(obj):
    if isfunction(obj):
        return obj.__name__
    elif isinstance(obj, partial):
        return obj.func.__name__
    else:
        return obj.__class__.__name__

def capture_artifact(function: T) -> T:
    @wraps(function)
    def artifact_write(*args, **kwargs):
        artifact = {
            'type': 'core_api_call',
            'query': _func_name(function),
            'args': list(args) + (
                list(function.args) if isinstance(function, partial) else []),
            'kwargs': {**kwargs,
                **(function.keywords if isinstance(function, partial) else {})
            },
        }
        stdout_ = StringIO()
        stderr_ = StringIO()
        try:
            with redirect_stdout(stdout_), redirect_stderr(stderr_):
                artifact['result'] = function(*args, **kwargs)
        except HTTPError as e:
            artifact['error'] = str(e)
        finally:
            artifact['stdout'] = stdout_.getvalue()
            artifact['stderr'] = stderr_.getvalue()

        stdout_.close()
        stderr_.close()

        sys.stdout.write(artifact['stdout'])
        sys.stderr.write(artifact['stderr'])
        manager.write_artifact(artifact)
        return artifact['result']

    return artifact_write



import rich
from malevich_space.ops import SpaceOps
from malevich_space.schema import SpaceSetup
from pydantic import ValidationError

from malevich.manifest import ManifestManager


def get_user_on_space() -> None:
    manf = ManifestManager()
    if (setup := manf.query('space', resolve_secrets=True)) is None:
        rich.print("You are not connect to Malevich Space. Run `malevich space login`")

    if not isinstance(setup, dict):
        rich.print(
            "Manifest is invalid. `space:` should be a valid mapping, "
            f"but it is {type(setup)}"
        )
        return


    try:
        setup = SpaceSetup(**setup)
    except ValidationError as ve:
        rich.print(
            "Manifest is invalid. `space:` does not comply to the schema, "
            + ', '.join([e.msg for e in ve.errors()])
        )
        return
    try:
        SpaceOps(setup)
    except Exception as e:
        rich.print(f"Could not login: {type(e)}({e}). Run `malevich space login`")
        return
    rich.print(f"You are [yellow]{setup.username}[/yellow] at "
               f"{setup.api_url.replace('api.', 'space.', 1)}")
    rich.print(
        f"Active manifest is [i][bright_black] { manf.path } [/i][/bright_black]"
    )



from typing import Any
from .result import CoreResult, CoreResultPayload


class FirstValidResultProxy(CoreResult):
    def __init__(self, results: list[CoreResult]):
        self._results = results
        self._valid: int = None


    def __getattribute__(self, name: str) -> Any:
        if self._valid is not None:
            import malevich_coretools


    @cache
    def get(self) -> list[CoreResultPayload]:
        if self._valid is not None:
            return
           

    @cache
    def get_df(self) -> pd.DataFrame:
        """Converts the result to a single data frame

        It only works if the result is a single collection,
        otherwise it raises an error. Check the number of
        elements using :attr:`num_elements` property and use only if
        it is equal to 1

        Returns:
            DataFrame: The result as a data frame

        Raises:
            NotImplementedError: If the result is not a single collection
        """
        if result := self.get():
            if len(result) == 1:
                if result[0].is_collection():
                    return result[0].data
                elif result[0].is_asset():
                    return NotImplementedError(
                        "Cannot return a single DataFrame from an asset. "
                        "Please use `get_binary` or `get_binary_dir` instead"
                    )
            else:
                raise NotImplementedError(
                    "Cannot return a single DataFrame from multiple results. "
                    "Please use `get_dfs` instead and check the number of results "
                    "using `num_elements` property"
                )
        else:
            warnings.warn(f"No results found for {self.core_group_name}")
            return pd.DataFrame()

    @cache
    def get_dfs(self) -> list[pd.DataFrame]:
        """Converts the result to a list of data frames

        If some of the results are assets, they are ignored
        in the output.

        Returns:
            list[:class:`DataFrame`]: The result as a list of data frames
        """
        if result := self.get():
            results_ = []
            for res in result:
                if res.is_collection():
                    results_.append(res.data)
                elif res.is_asset():
                    continue
            return results_
        else:
            return []

    @cache
    def get_binary(self) -> bytes:
        """Retrieves asset binary data, if the result is file asset

        It only works if the result is a single file returned
        as asset. Otherwise it raises an error. Check the number of
        elements using :attr:`num_elements` property and use only if
        it is equal to 1

        Returns:
            bytes: The binary data of the asset

        Raises:
            NotImplementedError: If the result is not a single asset
        """
        if result := self.get():
            if len(result) == 1:
                if result[0].is_collection():
                    raise NotImplementedError(
                        "Cannot return a single binary from a collection. "
                        "Please use `get_df` or `get_dfs` instead"
                    )
                elif result[0].is_asset():
                    return result[0].data
            else:
                raise NotImplementedError(
                    "Cannot return a single binary from multiple results. "
                    "Please use `get_binary_dir` instead and "
                    "check the number of results "
                    "using `num_elements` property"
                )
        else:
            warnings.warn(f"No results found for {self.core_group_name}")
            return b""

    @cache
    def get_binary_dir(self) -> dict[str, bytes]:
        """Retrieves files from assets

        Non-assets are ignored in the output

        Returns:
            dict[str, bytes]: Dict of file names and their binary data
        """
        if result := self.get():
            results_ = {}
            for res in result:
                if res.is_collection():
                    raise NotImplementedError(
                        "Cannot return a binary directory from a collection. "
                        "Please use `get_df` or `get_dfs` instead"
                    )
                elif res.is_asset():
                    if res.is_composite_asset():
                        results_.update(res.data)
                    else:
                        results_[res.get_path()] = res.data
            return results_
        else:
            warnings.warn(f"No results found for {self.core_group_name}")
            return {}

    @cache
    def get_document(
        self,
        model: type[DocumentModelType] | None = None
    ) -> dict | DocumentModelType:
        """Retrieves the document of the result

        Returns:
            dict: The document of the result
        """
        if result := self.get():
            # if result[index].is_document:
            #     return result[index].data if model is None else model(
            #         **result[index].data
            #     )
            # else:
            #     what = "asset" if result[index].is_asset() else "collection"
            #     raise NotImplementedError(
            #         "Cannot return a document from a non-document result. "
            #         f"Result at index {index} is {what}"
            #     )
            data_ = result[0].data.to_dict(orient='records')[0]
            return model(**data_) if model else data_

    @cache
    def get_documents(
        self,
        model: type[DocumentModelType] | None = None
    ) -> list[dict] | list[DocumentModelType]:
        if result := self.get():
            # if result[index].is_document:
            #     return result[index].data if model is None else model(
            #         **result[index].data
            #     )
            # else:
            #     what = "asset" if result[index].is_asset() else "collection"
            #     raise NotImplementedError(
            #         "Cannot return a document from a non-document result. "
            #         f"Result at index {index} is {what}"
            #     )
            data_ = result[0].data.to_dict(orient='records')
            if model:
                return [model(**d) for d in data_]
            return data_




import importlib.util
import pickle

import typer

from malevich.interpreter import SpaceInterpreter
from malevich.manifest import ManifestManager
from malevich.models import FlowFunction, PromisedTask

manf = ManifestManager()

flow = typer.Typer(
    name='flow',
    help="Manage flows uploading them without running",
)


@flow.command("export", help="Upload a flow to Malevich Core")
def export_flow(
    python_file: str = typer.Argument(...,
                                      help="Path to a Python file with a flow"),
    prefix: str = typer.Option("", help="Prefix for the flow name"),
) -> None:
    # Importing python file with a flow
    spec = importlib.util.spec_from_file_location("flow", python_file)

    if spec is None:
        raise typer.Exit(1)

    flow = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(flow)

    fn = []
    # Exporting flow

    for obj in flow.__dict__.values():
        if isinstance(obj, FlowFunction):
            fn.append(obj)

    if len(fn) == 0:
        typer.echo(f"There is no flows in {python_file}. Create one with @flow decorator")  # noqa: E501

    for f in fn:
        try:
            task: PromisedTask = f()
        except Exception as e:
            typer.echo(f"Skipping {f.reverse_id}. Most probably you are lacking "
                       "some modules installed, or the function requires arguments. "
                       f"Error: {e}")
        task.tree

        # Uploading secrets
        # TODO: upload secrets

        # Dumping flow to json
        flow_data = {
            "tree": task.tree,
            "apps": [next(iter(app.keys())) for app in manf.query("dependencies")]
        }
        flow_bytes = pickle.dumps(flow_data)

        with open(f"{prefix}{f.reverse_id}.malevichflow", "wb") as fl:
            fl.write(flow_bytes)

        typer.echo(
            f"Flow {f.reverse_id} was exported to {prefix}{f.reverse_id}.malevichflow"
        )


@flow.command("upload", help="Upload a flow to Malevich Space")
def upload_flow(
    path: str = typer.Argument(
        ...,
        help="Path to a file with a flow"
    ),
) -> None:
    # Uploading flow
    with open(path, "rb") as fl:
        flow_bytes = fl.read()
    data = pickle.loads(flow_bytes)
    tree = data["tree"]
    apps = data["apps"]
    for app in apps:
        importlib.import_module(f"malevich.{app}")

    typer.echo('Tree is restored successfully. Trying to interpret'
               ' with default Space Interpreter')

    interpreter = SpaceInterpreter()
    interpreter.interpret(tree)
    typer.echo(f"Flow {tree.name} was uploaded")



import enum
import hashlib
import importlib
import json
import os
import pickle
import uuid
import warnings
from copy import deepcopy
from typing import Any, Iterable, Literal, Optional, Type

import malevich_coretools as core
import pandas as pd
from malevich_space.schema import ComponentSchema
from pydantic import BaseModel, ValidationError

from malevich._autoflow.tracer import traced, tracedLike
from malevich._core.ops import (
    batch_upload_collections,
)
from malevich._utility import IgnoreCoreLogs, LogLevel, cout, upload_zip_asset
from ...nodes.morph import MorphNode
from ...._utility.cache.manager import CacheManager
from malevich.models import (
    Action,
    AssetNode,
    BaseNode,
    Collection,
    CollectionNode,
    CoreInjectable,
    CoreInterpreterState,
    CoreLocalDFResult,
    CoreResult,
    MetaEndpoint,
    OperationNode,
    TreeNode,
    VerbosityLevel,
)
from malevich.types import FlowOutput

from ...._utility.package import PackageManager
from ...exceptions import NoPipelineFoundError, NoTaskToConnectError
from ...overrides import AssetOverride, CollectionOverride, DocumentOverride, Override
from ..base import BaseTask


class BootError(Exception):
    ...


class PrepareStages(enum.Enum):
    BUILD = 0b01
    BOOT = 0b10
    ALL = 0b11


class CoreTaskStage(enum.Enum):
    ONLINE = 0
    BUILT = 1
    NO_TASK = 3


class CoreTaskState(CoreInterpreterState):
    unique_task_hash: str | None = None
    config: core.Cfg | None = None
    config_id: str | None = None
    pipeline_id: str | None = None


class CoreTask(BaseTask):
    """Represents a task on Malevich Core.

    Provides a user-friendly interfaces to interact with the task
    using Malevich Core API.
    """

    supports_conditional_output = True

    @staticmethod
    def load(object_bytes: bytes) -> 'CoreTask':
        return pickle.loads(object_bytes)

    def __init__(
        self,
        state: CoreInterpreterState,
        component: ComponentSchema | None = None
    ) -> None:

        if state is None:
            raise Exception("CoreTask requires a self.state to be passed. ")

        self.state = CoreTaskState()
        for key in CoreInterpreterState.model_fields.keys():
            setattr(self.state, key, getattr(state, key))

        self.run_id = None
        self.component = component

        self._returned = None

        CacheManager().core.write_entry(
            self.get_pipeline().model_dump_json(indent=4),
            entry_name=self.get_pipeline_hash() + '.json',
            entry_group='pipelines',
            force_overwrite=True
        )

    def get_active_tasks(self) -> list[str]:
        return self.state.service.run.active.list().ids

    def get_stage(self) -> CoreTaskStage:
        if self.state.pipeline_id is not None:
            try:
                with IgnoreCoreLogs():
                    runs = self.get_active_tasks()
                    if self.state.pipeline_id in runs:
                        return CoreTaskStage.ONLINE
            except Exception:
                pass

        if self.state.unique_task_hash is not None:
            try:
                with IgnoreCoreLogs():
                    self.state.service.pipeline.id(
                        self.state.unique_task_hash
                    ).get()
                return CoreTaskStage.BUILT
            except Exception:
                pass

        return CoreTaskStage.NO_TASK

    def get_stage_class(self) -> Type[CoreTaskStage]:
        return CoreTaskStage

    def _configure(
        self,
        operation: str,
        # Configurable parameters
        platform: str = 'base',
        platform_settings: Optional[dict[str, Any]] = None,
        # Rest of the parameters for compatibility
        **kwargs
    ) -> None:
        """internal"""
        assert platform in [
            'base', 'vast'], f"Platform {platform} is not supported. "
        cout(
            message=f"Configure {operation}: platform={platform}, platform_settings={platform_settings}",  # noqa: E501
            action=Action.Interpretation,
            verbosity=VerbosityLevel.OnlyStatus,
        )

        self.state.processors[operation].platform = platform
        if platform_settings:
            self.state.processors[operation].platformSettings = platform_settings

    def configure(
        self,
        *operations: str,
        # Configurable parameters
        platform: str = 'base',
        platform_settings: Optional[dict[str, Any]] = None,
        # Rest of the parameters for compatibility
        **kwargs
    ) -> None:
        """Configures the operation on Malevich Core

        Available configurations
        ------------------------

        Platform
        ++++++++

        The platform determines the environment in which the operation will be executed.
        We support two platforms: `base` and `vast`. The `base` platform is the default
        Malevich Core cluster. The `vast` platform is a on-demand GPU cluster that can
        be utilized for GPU-intensive operations.

        To configure the platform, use the `platform` parameter. The default value is
        `base`. To use the `vast` platform, set the value to `vast` and configure
        it with `malevich.core_api.vast_settings`.
        """
        for o in operations:
            self._configure(
                o,
                platform=platform,
                platform_settings=platform_settings,
                **kwargs
            )

    def get_pipeline(self, with_hash=False) -> core.Pipeline:
        pipeline = core.Pipeline(
            pipelineId='',
            processors=self.state.processors,
            conditions=self.state.conditions,
            results=self.state.results
        )
        if with_hash:
            pipeline.pipelineId = self.get_pipeline_hash()

        return pipeline

    def get_pipeline_hash(self) -> str:
        return hashlib.sha256(
           self.get_pipeline(with_hash=False).model_dump_json().encode() 
        ).hexdigest()

    async def prepare(
        self,
        stage: PrepareStages = PrepareStages.ALL,
        *args,
        **kwargs
    ) -> tuple[str, str]:
        """Prepares the task to be executed on Malevich Core

        The method is divided into two stages: build and boot.
        During build stage all the necessary components are send
        to the Core. During boot stage the task is actually
        deployed and gets ready for accepting runs.

        Args:
            - stage (PrepareStages, optional): The stage to be executed.
                Defaults to PrepareStages.ALL.
            - *args (Any, optional):
                Positional arguments to be passed to the :func`malevich.core_api.task_prepare`
                function.
            - **kwargs (Any, optional):
                Keyword arguments to be passed to the :func`malevich.core_api.task_prepare`
                function.

        Returns:
            tuple[str, str]: The task and operation IDs within Malevich Core. Some
                of them may be None if stage is specified.
        """  # noqa: E501
        cout(
            action=Action.Preparation,
            message="Task is being prepared for execution. It may take a while",
            verbosity=VerbosityLevel.OnlyStatus,
            level=LogLevel.Info
        )
        service = self.state.service
        for node in self.state.collection_nodes.values():
            collection = node.collection
            collection.core_id = (
                service.collection.name(collection.magic())
                .update_or_create(collection.collection_data)
            )

        for node in self.state.asset_nodes.values():
            if node.core_path is not None:
                with IgnoreCoreLogs():
                    try:
                        try:
                            files = service.asset.path(node.core_path).list(
                                recursive=True
                            ).files
                        except Exception as fe:
                            try:
                                files = service.asset.path(node.core_path).get()
                            except Exception as e:
                                raise fe from e



                        if node.real_path is not None:
                            if isinstance(files, bytes):
                                if isinstance(node.real_path, str):
                                    if os.path.getsize(node.real_path) != len(files):
                                        raise FileNotFoundError(
                                            f'{node.real_path} size mismatch'
                                        )
                                elif isinstance(node.real_path, list) and len(node.real_path) == 1:  # noqa: E501
                                    if os.path.getsize(node.real_path[0]) != len(files):
                                        raise FileNotFoundError(
                                            f'{node.real_path} size mismatch'
                                        )
                                else:
                                    raise FileNotFoundError(
                                        "Multiple files specified, but core asset is a single file"  # noqa: E501
                                    )
                            else:
                                for file in node.real_path:
                                    if file not in files:
                                        raise FileNotFoundError(f'{file} missing')
                                    if os.path.getsize(file) != files[file]:
                                        raise FileNotFoundError(
                                            f'{file} size mismatch')
                    except Exception as e:
                        if isinstance(e, FileNotFoundError):
                            message = e.strerror
                        else:
                            message = 'Failed to fetch asset'

                        service.asset.path(node.core_path).create(
                            file=node.real_path if isinstance(node.real_path, str) else None,  # noqa: E501
                            files=node.real_path if isinstance(node.real_path, list) else None,  # noqa: E501
                        )

                        cout(
                            action=Action.Preparation,
                            message=f"Asset {node.name} updated. {message}",
                            verbosity=VerbosityLevel.AllSteps,
                            level=LogLevel.Debug
                        )

                    else:
                        cout(
                            action=Action.Preparation,
                            message=f"Asset {node.name} fully matched with the Core",
                            verbosity=VerbosityLevel.AllSteps,
                            level=LogLevel.Debug
                        )

        for node in self.state.document_nodes.values():
            try:
                ref = service.document.name(node.magic())
                with IgnoreCoreLogs():
                    node.core_id = ref.get().id
                cout(
                    action=Action.Preparation,
                    message=f"Document {node.reverse_id} is already on Core. {node.magic()}",  # noqa: E501
                    verbosity=VerbosityLevel.AllSteps,
                    level=LogLevel.Debug
                )
            except Exception:
                node.core_id = ref.create(data=node.dump_document_json())

                cout(
                    action=Action.Preparation,
                    message=f"Document {node.reverse_id} uploaded. {node.magic()}",
                    verbosity=VerbosityLevel.AllSteps,
                    level=LogLevel.Debug
                )

        if not self.state.config:
            config = core.Cfg(
                collections={
                    v.collection.collection_id: v.collection.core_id
                    for k, v in self.state.collection_nodes.items()
                }
            )

            for an in self.state.asset_nodes.values():
                config.collections = {
                    **config.collections,
                    an.name: an.get_core_path(),
                }

            for dn in self.state.document_nodes.values():
                config.collections = {
                    **config.collections,
                    dn.reverse_id: dn.get_core_path(),
                }

            self.state.config = config

        if stage.value & PrepareStages.BUILD.value:
            if not self.state.unique_task_hash:
                self.state.unique_task_hash = self.get_pipeline_hash()

            self.state.config_id = self.state.unique_task_hash
            try:
                pref = service.pipeline.id(self.state.unique_task_hash)
                with IgnoreCoreLogs():
                    self.state.pipeline_id = pref.get().pipelineId
                    cout(
                        action=Action.Preparation,
                        message=f"Pipeline {self.state.unique_task_hash} found.",
                        verbosity=VerbosityLevel.AllSteps,
                        level=LogLevel.Debug
                    )

            except Exception:
                self.state.pipeline_id = pref.create(
                    processors=self.state.processors,
                    conditions=self.state.conditions or None,
                    results=self.state.results,
                )

                cout(
                    action=Action.Preparation,
                    message=f"Pipeline {self.state.unique_task_hash} created.",
                    verbosity=VerbosityLevel.AllSteps,
                    level=LogLevel.Debug
                )

        if stage.value & PrepareStages.BOOT.value:
            if self.state.pipeline_id is None:
                raise BootError(
                    "Failed to boot: no pipeline found. "
                    "Try `.prepare(stage=PrepareStages.BUILD)` or reinterpret the task"
                )
            try:
                service.cfg.name(self.state.config_id).update_or_create(
                    cfg_id=self.state.config_id,
                    cfg=self.state.config,
                )
                piperef = service.pipeline.id(self.state.unique_task_hash)
                self.state.params.operation_id = piperef.prepare(
                    cfg_id=self.state.unique_task_hash,
                    *args,
                    **kwargs
                ).operationId
            except (Exception, KeyboardInterrupt) as e:
                try:
                    service.run.operation_id(self.state.params.operation_id).stop()
                except Exception:
                    pass
                raise e

        return self.state.unique_task_hash, self.state.params.operation_id

    def _prepare_collection_overrides(
        self,
        injectables: list[CoreInjectable],
        overrides: dict[str, CollectionOverride]
    ) -> dict[str, str]:
        collections = [
            Collection(
                collection_id=f'core_interpreter_override_{k}_{self.run_id}',
                collection_data=v.data,
                persistent=False
            ) for k, v in overrides.items()
        ]

        refs = [
            self.state.service.collection.name(collection.magic())
            for collection in collections
        ]

        core_ids = [
            ref.update_or_create(data=col.collection_data)
            for ref, col in zip(refs, collections)
        ]

        key_to_core_id = {
            k: core_id
            for k, core_id in zip(overrides.keys(), core_ids)
        }

        return {
            injectable.get_inject_key():
            key_to_core_id[injectable.get_inject_key()]
            for injectable in injectables
            if injectable.get_inject_key() in overrides
        }

    def _prepare_asset_overrides(
        self,
        injectables: list[CoreInjectable],
        overrides: dict[str, AssetOverride]
    ) -> dict[str, str]:
        real_overrides = {}

        for k, v in overrides.items():
            if v.path is not None:
                if v.file is not None or v.files is not None:
                    self.state.service.asset.path(v.path).create(
                        file=v.file,
                        files=v.files
                    )
                else:
                    real_overrides[k] = v.path

        return {
            injectable.get_inject_key():
            real_overrides[injectable.get_inject_key()]
            for injectable in injectables
            if injectable.get_inject_key() in real_overrides
        }

    def _prepare_document_overrides(
        self,
        injectables: list[CoreInjectable],
        overrides: dict[str, DocumentOverride]
    ) -> dict[str, str]:
        real_overrides = {}

        for k, v in overrides.items():
            name = hashlib.sha256(v.data.model_dump_json().encode()).hexdigest() + '_override'  # noqa: E501
            id = self.state.service.document.name(name).update_or_create(
                data=v.data.model_dump_json(),
            )
            real_overrides[k] = f'#{id}'

        return {
            injectable.get_inject_key():
            real_overrides[injectable.get_inject_key()]
            for injectable in injectables
            if injectable.get_inject_key() in real_overrides
        }

    def _validate_proc_cfg_ext(
        self,
        package_id: str,
        processor_id: str,
        config_extension,
    ):
        """internal"""
        from malevich._meta.decor import ProcessorFunction
        try:
            PackageManager().get_package(package_id)
            module = importlib.import_module(f'malevich.{package_id}')
        except Exception:
            return None
        if not hasattr(module, processor_id):
            return None
        proc_stub = getattr(module, processor_id)
        if not isinstance(proc_stub, ProcessorFunction):
            return None

        if isinstance(proc_stub.config, BaseModel) and isinstance(config_extension, BaseModel):  # noqa: E501
            return type(proc_stub.config) == type(config_extension)  # noqa: E721
        else:
            return True

    def _get_config_model(self, package_id: str, processor_id: str) -> BaseModel | None:
        from malevich._meta.decor import ProcessorFunction

        try:
            module = importlib.import_module(f'malevich.{package_id}')
        except ImportError:
            return None

        proc_stub = getattr(module, processor_id)
        if isinstance(proc_stub, ProcessorFunction):
            if issubclass(proc_stub.config, BaseModel):
                return proc_stub.config
        return None

    def _validate_extension(
        self,
        config_extension: dict[str, dict[str, Any] | BaseModel],
    ) -> str:
        app_cfg_extensions = {}
        print([*self.state.operation_nodes.values()])
        for alias, extension in config_extension.items():
            for x in self.state.operation_nodes.values():
                if isinstance(x, OperationNode) and x.alias == alias:
                    config_model: BaseModel | None = self._get_config_model(
                        x.package_id,
                        x.processor_id
                    )
                    if not isinstance(extension, BaseModel) and not isinstance(extension, dict):  # noqa: E501
                        _expected = "dictionary"
                        if config_model is not None and isinstance(config_model, BaseModel):  # noqa: E501
                            _expected += f'or {type(config_model).__name__}'

                        raise ValueError(
                            "Invalid type for config extension. "
                            f"Expected {_expected}, but found {type(extension).__name__} "  # noqa: E501
                            f"for alias {x.alias}, processor {x.processor_id} and "
                            f"package {x.package_id}."
                        )

                    if config_model is not None:
                        try:
                            config_model(**{
                                **x.config,
                                **(extension if isinstance(extension, dict)
                                    else extension.model_dump()
                                    )
                            })
                        except ValidationError as e:
                            raise ValueError(
                                "Failed to extend the configuration "
                                f"for processor {x.processor_id} with alias"
                                f" {x.alias}. New configuration do not comprise "
                                "to the schema of the processor of the config. "
                                "See validation errors above."
                            ) from e
                        if (
                            config_model is not None
                            and issubclass(config_model, BaseModel)
                            and not issubclass(config_model,  type(extension))
                        ):
                            raise ValueError(
                                "Failed to extend the configuration "
                                f"for processor {x.processor_id} with alias"
                                f" {x.alias}. Expected {config_model.__name__}, "
                                f"but the configuration extenstion is {type(extension).__name__}"  # noqa: E501
                            )

                    if isinstance(extension, BaseModel):
                        extension_json = extension.model_dump_json()
                    elif isinstance(extension, dict):
                        extension_json = json.dumps(extension)
                    else:
                        # ok, validates before
                        pass
                    app_cfg_extensions['$' + x.alias] = extension_json

        return app_cfg_extensions

    def _compile_overrides(
        self,
        override: dict[str, Override]
    ):
        collection_overrides = {
                k: v for k, v in override.items()
                if isinstance(v, CollectionOverride)
            }
        asset_overrides = {
            k: v for k, v in override.items()
            if isinstance(v, AssetOverride)
        }

        document_overrides = {
            k: v for k, v in override.items()
            if isinstance(v, DocumentOverride)
        }

        injectables = self.get_injectables()
        real_overrides = {
            **self._prepare_collection_overrides(
                injectables,
                collection_overrides
            ),
            **self._prepare_asset_overrides(
                injectables,
                asset_overrides
            ),
            **self._prepare_document_overrides(
                injectables,
                document_overrides
            )
        }

        for k, v in real_overrides.items():
            cout(
                message=f"Override {k} with {v}",
                action=Action.Run,
                verbosity=VerbosityLevel.AllSteps,
                level=LogLevel.Debug
            )

        return real_overrides


    async def run(
        self,
        override: dict[str, Override] | None = None,
        config_extension: dict[str, dict[str, Any] | BaseModel] | None = None,
        run_id: Optional[str] = None,
        detached: bool = False,
        stop_on_error: bool = False,
        stop_on_interrupt: bool = False,
        *args,
        **kwargs
    ) -> str:
        """Runs the task on Malevich Core

        You can supply new data or new configuration each run.

        New data is supplied using overrides. An override is a dictionary
        that maps the alias of the data node (e.g. collection, asset, document)
        to the new value via Override interface.

        Collections overrides are created using :class:`CollectionOverride`
        class.

        Args:
            - overrides (dict[str, str], optional):
                Collection overrides within Malevich Core.
                The dict in the form:
            - run_id (str, optional): The ID of the run. Defaults to None.
            - detached (bool, optional): Whether to wait for the task to finish. Defaults
                to False.
            - *args (Any, optional): Positional arguments to be passed to the
                :func:`malevich.core_api.task_run` function.
            - **kwargs (Any, optional): Keyword arguments to be passed to the
                :func:`malevich.core_api.task_run` function.

        """  # noqa: E501
        cout(
            message="Task is being executed on Core. It may take a while",
            action=Action.Run,
            verbosity=VerbosityLevel.OnlyStatus,
        )

        if "operation_id" not in self.state.params:
            raise Exception("Attempt to run a task which is not prepared. "
                            "Please, run `.prepare()` first.")

        if override:
            real_overrides = self._compile_overrides(override)
        else:
            real_overrides = {}

        self.run_id = run_id or uuid.uuid4().hex

        app_cfg_extensions = {}
        if config_extension:
            app_cfg_extensions = self._validate_extension(config_extension)

        tref = self.state.service.run.operation_id(
            self.state.params.operation_id
        )
        try:
            if real_overrides or app_cfg_extensions:
                new_config = self.state.config.model_copy(deep=True)
                new_config.collections = {
                    **self.state.config.collections,
                    **real_overrides,
                }
                print(self.state.config.collections, real_overrides)
                new_config.app_cfg_extension = app_cfg_extensions
                new_config_id = self.state.config_id + \
                    '_' + uuid.uuid4().hex[:6]

                self.state.service.cfg.name(new_config_id).update_or_create(
                    cfg_id=new_config_id,
                    cfg=new_config,
                )
                tref.run(
                    cfg_id=new_config_id,
                    wait=not detached,
                    run_id=self.run_id,
                    *args,
                    **kwargs
                )
            else:
                tref.run(
                    *args,
                    run_id=self.run_id,
                    wait=not detached,
                    **kwargs
                )
        except Exception as e:
            if stop_on_error:
                await self.stop()
            raise e
        except KeyboardInterrupt:
            if stop_on_interrupt:
                await self.stop()
            raise
        return self.run_id

    async def stop(
        self,
        *args,
        **kwargs
    ) -> None:
        """Stops the task preventing it from further execution

        Args:
            - *args (Any, optional): Positional arguments to be passed to the
                :func:`malevich.core_api.task_stop` function.
            - **kwargs (Any, optional): Keyword arguments to be passed to the
                :func:`malevich.core_api.task_stop` function.
        """
        if "operation_id" not in self.state.params:
            raise Exception("Attempt to run a task which is not prepared. "
                            "Please, run `.prepare()` first.")
        self.state.service.run.operation_id(
            self.state.params.operation_id
        ).stop(*args, **kwargs)

    async def results(
        self,
        # returned: Iterable[traced[BaseNode]] | traced[BaseNode] | None,
        run_id: Optional[str] = None,
        # For compatibility with other interpreters
        *args,
        **kwargs
    ) -> list[CoreResult | CoreLocalDFResult]:
        cout(message="Task results are being fetched from Core",
             action=Action.Results)
        if self.state.params.operation_id is None:
            raise Exception("Attempt to run a task which is not prepared. "
                            "Please, run `.prepare()` first.")

        returned = self._returned
        if not self._returned:
            return None

        demorphed_returned = []
        for i in range(len(returned)):
            if isinstance(returned[i][1].owner, MorphNode):
                for morph_conditions, node in returned[i][1].owner.members:
                    demorphed_returned.append(({
                        **(returned[i][0] or {}),
                        **(morph_conditions or {})
                        }, node,
                    ))
            else:
                demorphed_returned.append(returned[i])

        returned = demorphed_returned

        if not run_id:
            run_id = self.run_id

        logs = core.logs(
            self.state.params.operation_id,
            run_id=run_id,
            auth=self.state.params.core_auth,
            conn_url=self.state.params.core_host,
        )
        no_conditions_return = None
        final_result = None

        condition_map = {}
        for alias, info in logs.pipeline.conditions.items():
            condition_map[alias] =  info[max(info.keys())]

        for condition, return_map in returned:
            if condition is None:
                no_conditions_return = return_map
            else:
                matched = True
                for node, value in condition.items():
                    matched &= condition_map[node.alias] == value
                if matched:
                    final_result = return_map

        if not condition_map:
            final_result = no_conditions_return

        returned = final_result

        if not isinstance(returned, list):
            returned = [returned]

        def _deflat(li: list[BaseNode]) -> list[BaseNode]:
            temp_returned = []
            for r in li:
                if isinstance(r, TreeNode):
                    temp_returned.extend(
                        _deflat(r.results)
                    )
                else:
                    temp_returned.append(r)
            return temp_returned

        returned = _deflat(returned)
        returned = [x.owner for x in returned]
        results = []
        for node in returned:
            if isinstance(node, CollectionNode):
                results.append(
                    CoreLocalDFResult(
                        coll=node.collection,
                        conn_url=self.state.params.core_host,
                        auth=self.state.params.core_auth,
                    )
                )
            elif isinstance(node, OperationNode):
                results.append(CoreResult(
                    core_group_name=node.alias,
                    core_operation_id=self.state.params.operation_id,
                    core_run_id=run_id,
                    auth=self.state.params.core_auth,
                    conn_url=self.state.params.core_host,
                ))
            elif isinstance(node, AssetNode):
                results.append(CoreResult(
                    core_group_name=node.alias,
                    core_operation_id=self.state.params.operation_id,
                    core_run_id=run_id,
                    conn_url=self.state.params.core_host
                ))
            else:
                warnings.warn(
                    f"Cannot interpret {type(node)} as a result."
                )

        return results

    def get_injectables(
        self,
    ) -> list[CoreInjectable]:
        """Retrieves a list of injections available for Malevich Core

        CoreInjectable is a representation of possible injection.
        To inject a collection into the task, you need to create
        a new collection using Malevich Core API.

        Use `get_inject_data()` as the key for `overrides` in the `run` method.

        Use `get_inject_key()` to match the alias of overriden collection.

        Returns:
            list[CoreInjectable]: A list of injections available for Malevich Core
        """
        injectables = []
        injectables.extend([
            CoreInjectable(
                alias=node.alias,
                node=node,
                collection_id=node.collection.collection_id,
            ) for node in self.state.collection_nodes.values()])

        injectables.extend([
            CoreInjectable(
                alias=node.alias,
                node=node,
                collection_id='$' + (node.core_path or node.real_path)
            ) for node in self.state.asset_nodes.values()])

        injectables.extend([
            CoreInjectable(
                alias=node.alias,
                node=node,
                collection_id='#' + (node.core_id or node.reverse_id)
            ) for node in self.state.document_nodes.values()])

        return injectables

    def get_operations(self, *args, **kwargs) -> list[str]:
        return [
            x.alias
            for x in self.state.operation_nodes.values()
        ]

    def commit_returned(
        self, returned: FlowOutput | dict[dict[str, bool], FlowOutput]
    ) -> None:
        self._returned = returned

    def dump(self) -> bytes:
        return pickle.dumps(self)

    def get_interpreted_task(self) -> BaseTask:
        return self

    def interpret(self, interpreter = None) -> None:
        raise Exception(
            "Trying to re-interpret task of type `CoreTask`. You can only interpret "
            "`PromisedTask`"
        )

    @property
    def tree(self) -> None:
        raise Exception(
            "Trying to access a tree of type `CoreTask`. Get tree from `PromisedTask`"
            " instead."
        )

    def connect(
        self,
        unique_task_hash: str | None = None,
        only_fetch: bool = False,
    ) -> 'CoreTask':
        unique_task_hash = unique_task_hash or self.get_pipeline_hash()
        try:
            pipeline = self.state.service.pipeline.id(unique_task_hash).get()
        except Exception:
            raise NoPipelineFoundError(unique_task_hash)

        self.state.processors = pipeline.processors
        self.state.results = pipeline.results
        self.state.conditions = pipeline.conditions
        self.state.unique_task_hash = unique_task_hash
        self.state.config = core.Cfg()
        self.state.config_id = unique_task_hash

        json_cfg = json.loads(
            self.state.service.cfg.id(unique_task_hash).get().data
        )

        for key, value in json_cfg.items():
            setattr(self.state.config, key, value)

        if self.state.operation_nodes:
            node_results = [
                tracedLike(self.state.operation_nodes[x])
                for x in pipeline.results.keys()
            ]
        else:
            node_results = [
                tracedLike(OperationNode(alias=x, operation_id=''))
                for x in pipeline.results.keys()
            ]

        self.commit_returned(node_results)
        if not only_fetch:
            tasks = self.get_active_tasks()

            if tasks:
                self.state.params.operation_id = tasks[-1]
                cout(
                    Action.Attachment,
                    message='Connected to online task. ' + tasks[-1],
                    level=LogLevel.Info,
                    verbosity=VerbosityLevel.OnlyStatus
                )
            else:
                raise NoTaskToConnectError()

        return self

    def publish(
        self,
        capture_results: list[str] | Literal['all'] | Literal['last'] = 'last',
        enable_not_auth: bool = True,
        hash: str | None = None,
        *args,
        **kwargs
    ) -> MetaEndpoint:
        from malevich_coretools import create_endpoint, update_endpoint
        if self.get_stage() not in [CoreTaskStage.BUILT, CoreTaskStage.ONLINE]:
            self.prepare(stage=PrepareStages.BUILD)

        cfg = deepcopy(self.state.config)
        cfg_id = f'endpoint-config-{uuid.uuid4().hex}'

        self._create_cfg_safe(
            cfg_id=cfg_id,
            cfg=cfg
        )

        if 'prepare' not in kwargs:
            kwargs['prepare'] = True

        if not hash:
            _hash = create_endpoint(
                pipeline_id=self.state.pipeline_id,
                cfg_id=cfg_id,
                auth=self.state.params.core_auth,
                conn_url=self.state.params.core_host,
                enable_not_auth=enable_not_auth,
                description=(
                    self.component.description if self.component
                    else 'Auto-generated meta endpoint'
                ),
                *args,
                **kwargs
            )
        else:
            _hash = update_endpoint(
                hash,
                pipeline_id=self.state.pipeline_id,
                cfg_id=cfg_id,
                auth=self.state.params.core_auth,
                conn_url=self.state.params.core_host,
                enable_not_auth=enable_not_auth,
                description=(
                    self.component.description if self.component
                    else 'Auto-generated meta endpoint'
                ),
                *args,
                **kwargs
            )

        _endpoint = core.get_endpoint(
            hash=_hash,
            auth=self.state.params.core_auth,
            conn_url=self.state.params.core_host,
        )

        return MetaEndpoint(
            **_endpoint.model_dump(),
            conn_url=self.state.params.core_host
        )
